#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/spi.h> 
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#define SPI_BASE spi_base
#define SPI_DATA spi_base
#define SPI_STAT spi_base+1
#define SPI_CNTL spi_base+2

#define SPI_IDLE 3		/* idle state of MOSI and SCK is low */
#define SPI_NOCE 0xff		/* set all CEs high in this version */
#define SPI_CLK 2		/* bit 1 of command reg (AUTO_LF) */
#define SPI_DIN_LINE 0x80	/* bit 7 of status reg (BUSY) */

static	int (* spi_bit_driver) (int);
static	int	openflag;
static  unsigned char spi_ctl_bck;
static  int spi_base = 0x378;
unsigned int spi_major = 60; /* a local major, can be overwritten */

/* Some support routines first */

static int spi_transfer_byte (int databyte)
/* spi_transfer_byte sends one byte over the SPI bus and concurrently
   receives one (as specified in the SPI specs). The MSB is 
   transferred first.
   It expects the byte in the low byte of the int and returns the
   result in the low byte. The reason, that ints are used is to
   simplify the code.
*/
{
  int i;        /* the counter */
  int inval=0;  /* temp. var for the byte to read in */	
  int outval;   /* same for byte to send */

  for (i=0;i<8;i++){ 			/* loop 8 times */
    inval = inval << 1;                 /* shift new bit up */
    databyte = databyte << 1;           /* shift MSB of lower byte into HB */
    outval = (databyte & 0x100) >> 8;   /* mask bit, shift to LSB */
    inval=inval | (spi_bit_driver(outval)==0 ? 1 : 0 );	 /* transfer bit in MSB */  
  }/*end-for*/                          /* one byte is transferred */

  return inval ;	                   
}	



static int pha0_driver (int bit)

/* this is the driver for any microwire device.
It sends a one if "bit" is not zero and returns a non-zero value
if it read in a one. The bit set in the return value is not defined! If a zero
value is returned, a zero was read in and if a non-zero value is returned a one
was read in. 
*/
{
  int inxx ;	/* we must keep the bit for a while */

  outb_p(spi_ctl_bck ^ bit, SPI_CNTL); /* put out bit, deactivate SCK */
  inxx = inb_p(SPI_STAT) & SPI_DIN_LINE; /* read bit before clock pulse */ 
  outb_p( (spi_ctl_bck ^ bit) ^ SPI_CLK, SPI_CNTL );  /* activate SCK */
  return inxx;  
}




static int pha1_driver (int bit)

/* this is the driver for any SPI device.
It sends a one if "bit" is not zero and returns a non-zero value
if it read in a one. The bit set in the return value is not defined! If a zero
value is returned, a zero was read in and if a non-zero value is returned a one
was read in. 
*/
{
  int tmpo;
  tmpo = spi_ctl_bck ^ bit;		/* we need it more than once */
  outb_p(tmpo, SPI_CNTL);   		/* out bit, deactivate SCK */
  outb_p(tmpo ^ SPI_CLK, SPI_CNTL);   	/* out bit, activate SCK */
  outb_p(tmpo, SPI_CNTL);        	/* put out bit, deactivate SCK */
  outb_p(tmpo, SPI_CNTL);        	/* again, to let line settle */
  return inb_p(SPI_STAT) & SPI_DIN_LINE;/* read bit after clock pulse */
}


int select_proto (int proto)
{
  int retval = 0;

  switch(proto) {
    case SEL_SPI_CPOL_1:
      spi_bit_driver = pha1_driver;           /* set the correct bit driver */
      spi_ctl_bck = spi_ctl_bck & ~SPI_CLK ;  /* set clock high */
      outb_p(spi_ctl_bck, SPI_CNTL);		
      break;
    case SEL_SPI_CPOL_0:
      spi_bit_driver = pha1_driver;           /* set the correct bit driver */
      spi_ctl_bck = spi_ctl_bck | SPI_CLK ;   /* set clock low */
      outb_p(spi_ctl_bck, SPI_CNTL);		
      break;
    case SEL_UWIRE_CPOL_0:
      spi_bit_driver = pha0_driver;          /* set the correct bit driver */
      spi_ctl_bck = spi_ctl_bck | SPI_CLK ;  /* set clock low */
      outb_p(spi_ctl_bck, SPI_CNTL);		
      break;
    case SEL_UWIRE_CPOL_1:
      spi_bit_driver = pha0_driver;          /* set the correct bit driver */
      spi_ctl_bck = spi_ctl_bck & ~SPI_CLK ; /* set clock high */
      outb_p(spi_ctl_bck, SPI_CNTL);		
      break;
    default:
      retval = -EINVAL;
    }
    return retval;
}	



/* here come the driver functions now */
	
static int spi_reset(void)
{
  outb_p(SPI_NOCE, SPI_DATA);	/* set all CEs incative */
  outb_p(SPI_IDLE, SPI_CNTL);
  spi_ctl_bck = SPI_IDLE ;	/* remember the default idle state */
  return (inb_p(SPI_STAT));	/* return status, just for fun */
}

static int spi_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)

{
  int retval = 0;
  return retval;
}			



static int spi_write(struct inode * inode, struct file * file, 
			const char * buf, int count)

/* write data in buffer to spi bus and concurrently read in new.
The buffer must have the following structure in it:
<protocol specifier in the least low bits>
<CE settings to use at the start>
<data byte 1>
<additional data bytes>
<CE settings, that are to left behind>
	   
As there must be at least one data byte the minimum count is 4
(3 bytes overhead and at least one data byte)
The starting CE settings are applied just before the first byte
is transferred. The ending CE settings are applied after the last
data byte is transferred. In a one byte message the starting
setting adresses a chip and the ending deadresses it. When talking
to chips that need multibyte transfers with a pause inbetween, 
however, the start settings of the whole action adresses the chip
as usual, but the ending settings of any intermediate transfer
leaves the chip adressed.
	   
Each byte received supersedes the corresponding byte sent in the
buffer. Thus the buffer contains the received data bytes on
return from this function. As there are always exactly as many
bytes received as sent, there cannot be an overflow of the buffer.
*/
	   
{
  unsigned char *temp;                 /* our pointer to the buffer */
  int tmp;	
  int i;
  
  if(count<4) return -EINVAL;          /* must be at least 4 bytes in buf*/

  (char *) temp = buf;                 /* point to the user buffer */
  tmp = select_proto(get_fs_byte(temp));     /* set the protocol */
  if (tmp == -EINVAL) return -EINVAL ;
  
  temp++;                              /* advance to next byte */
  outb_p(get_fs_byte(temp), SPI_DATA); /* set the CEs */

  temp++;                              /* point to the first data byte */
  for (i=count-3; i>0; i--) {
    tmp=spi_transfer_byte(get_fs_byte(temp));
    put_fs_byte(tmp, temp);            /* write read in byte to buffer */
    temp++;                            /* advance to next byte */
  }/*end-for*/                         /* all data bytes transferred now */
  outb_p(spi_ctl_bck, SPI_CNTL); 	/* deactivate SCK */
  outb_p(get_fs_byte(temp), SPI_DATA); /* the ending CE settings */
  return count;                        /* we can always send all data */
}

static int spi_read(struct inode * inode, struct file * file, 
			char * buf, int count)
{ 
   spi_write(inode,file,(const char *)buf,count);
   return 0 ;
}


static int spi_lseek(struct inode * inode, struct file * file,
                     off_t offset, int origin)
{
  return -ESPIPE;                      /* the only action to do */
}

static int spi_open(struct inode * inode, struct file * file)
{
  if (openflag) return -EBUSY;         /* only one access at a time */
  openflag=1;                          /* mark it already opened */
  return 0;
}

static void spi_release(struct inode * inode, struct file * file)
{
  openflag=0;
}


/* static struct file_operations spi_fops  */

struct file_operations spi_fops = {
  spi_lseek,     /* prints an error message only */
  spi_read,      /* differs only in a cast from write routine */
  spi_write,
  NULL,          /* spi_readdir */
  NULL,          /* spi_select */
  spi_ioctl,	
  NULL,          /* spi_mmap */
  spi_open, 
  spi_release, 
  
};

void cleanup_module(void)
{
    if (spi_major > 0) unregister_chrdev(spi_major, "spi");
    release_region(SPI_BASE,3);
}



int init_module(void)
{
  int result = register_chrdev(spi_major,"spi",&spi_fops);
  if (result < 0) {
    printk("SPI: unable to get major %d for SPI bus \n", spi_major);
    return result;
  }/*end-if*/
  
  if (spi_major == 0) spi_major = result; /* here we got our major dynamically */

  /* reserve our port, but check first if free */
  if (check_region(SPI_BASE, 3)) {
    printk("SPI: port at adr 0x%04x already occupied\n", SPI_BASE);
    unregister_chrdev(spi_major, "spi");
    return result;
  }/*end-if*/

  /* write to port & read back to check */
  outb_p(0, SPI_DATA);
  if (inb_p(SPI_DATA) == 0) {
    spi_ctl_bck = SPI_IDLE ;
    spi_reset();
    request_region(SPI_BASE, 3, "spi");
    printk("SPI bus at 0x%04x \n", SPI_BASE);
    return 0;
  } else {
    cleanup_module();
    printk("SPI: no chips at 0x%04x \n", SPI_BASE);
    return -EINVAL ;
  }/*end-if*/
}

